module DocMachine::Container

  class PodmanEngine < AbstractContainerEngine
    # Pulls the specified Podman image
    def pull_image(image_tag : String) : Nil
      Log.info { "Pulling Podman image: #{image_tag}" }
      status = Process.run("podman", ["pull", image_tag], output: STDOUT, error: STDERR)
      if status.success?
        Log.info { "Successfully pulled Podman image: #{image_tag}" }
        true
      else
        Log.error { "Failed to pull Podman image: #{image_tag}" }
        false
      end
    end

    # Loads the Podman image from a local cache
    def load_image(cache_path : String) : Nil
      Log.info { "Loading Podman image from cache: #{cache_path}" }
      status = Process.run("podman", ["image", "load", "-i", cache_path], output: STDOUT, error: STDERR)
      if status.success?
        Log.info { "Successfully loaded Podman image from cache." }
      else
        Log.error { "Failed to load Podman image from cache." }
        raise LoadError.new("Failed to load Podman image from cache.")
      end
    end

    # Saves the Podman image to a local cache
    def save_image(image_tag : String, cache_path : String) : Nil
      Log.info { "Saving Podman image #{image_tag} to cache at #{cache_path}" }
      status = Process.run("podman", ["image", "save", image_tag, "-o", cache_path], output: STDOUT, error: STDERR)
      if status.success?
        Log.info { "Successfully saved Podman image to cache." }
      else
        Log.error { "Failed to save Podman image to cache." }
        raise SaveError.new("Failed to save Podman image to cache.")
      end
    end

    # Runs a Podman container with the given options
    def run_container(
      image_tag : String,
      action : String,
      docker_name : String,
      docker_opts : Array(String),
      enable_tty : Bool
    ) : Process
      Log.info { "Running Podman container: #{docker_name}" }
      
      # Construct the full Podman run command
      cmd = ["podman", "run"] + docker_opts + [image_tag, action]
      
      # Log the command for debugging
      Log.debug { "Podman run command: #{cmd.join(" ")}" }
      
      # Start the Podman container process
      process = Process.new("podman", ["run"] + docker_opts + [image_tag, action], input: STDIN, output: STDOUT, error: STDERR)
      
      Log.info { "Podman container #{docker_name} started." }
      
      process
    end

    # Kills a running Podman container by name
    def kill_container(container_name : String) : Nil
      Log.info { "Killing Podman container: #{container_name}" }
      status = Process.run("podman", ["kill", container_name], output: STDOUT, error: STDERR)
      if status.success?
        Log.info { "Successfully killed Podman container: #{container_name}" }
        true
      else
        Log.error { "Failed to kill Podman container: #{container_name}" }
        false
      end
    end

    # Removes a Podman container by name
    def remove_container(container_name : String) : Nil
      Log.info { "Removing Podman container: #{container_name}" }
      status = Process.run("podman", ["rm", container_name], output: STDOUT, error: STDERR)
      if status.success?
        Log.info { "Successfully removed Podman container: #{container_name}" }
        true
      else
        Log.error { "Failed to remove Podman container: #{container_name}" }
        false
      end
    end

    # Finds the container ID based on its name
    def find_container_id(name : String) : String
      Log.info { "Finding Podman container ID for name: #{name}" }
      output = IO::Memory.new
      status = Process.run("podman", ["ps", "-f", "name=#{name}", "-q"], output: output, error: STDERR)
      if status.success?
        container_id = output.to_s.strip
        Log.info { "Found Podman container ID: #{container_id}" }
        container_id
      else
        Log.error { "Failed to find Podman container ID for name: #{name}" }
        ""
      end
    end
  end

end
